import struct
from numbers import Number

from apogee.SerialClient import SerialClient
from apogee.modbus.ModbusSensorRegisterDict import ModbusSensorModel, Sensor_Registers

class ModbusClient(SerialClient):
    
    #region CONSTRUCTOR
    def __init__(self,
                 port: str,
                 baudrate: int = 9600,
                 stopbits: int = 1,
                 parity: str = 'N',
                 bytesize: int = 8,
                 protocol: str = 'RS-485',
                 timeout: int = 1,
                 skip_AC422_setup: bool = False 
                 ):
        """
        :param port: comport name of connected serial device
        :param baudrate: (optional) 115200, 57600, 38400, 1920, 9600
        :param stopbits: (optional) 0, 1, 2
        :param parity: (optional) 'N'one, 'O'dd, or 'E'ven 
        :param bytesize: (optional) 7, 8
        :param protocol: (optional) RS-232, RS-485
        :param timeout: (optional) No response timeout in seconds
        :param skip_AC422_setup: (optional) If True, will not configure AC-422 during initialization. May be used when not using AC-422.
        """
        self._validate_stopbits(stopbits)
        self._validate_parity(parity)
        self._validate_bytesize(bytesize)
        self._validate_protocol(protocol)

        self._stopbits = stopbits
        self._parity = parity
        self._bytesize = bytesize
        self._protocol = protocol

        super().__init__(port, baudrate, timeout)
        if not skip_AC422_setup:
            self._configure_modbus_interface_type()
    #endregion

    #region VALIDATION
    def _validate_stopbits(self, stop_bits):
        if stop_bits not in {1, 2}:
            raise ValueError(f"Invalid stopbits {stop_bits}. Must be 1 or 2.")

    def _validate_parity(self, parity):
        valid_parities = {'N', 'O', 'E'}
        if parity not in valid_parities:
            raise ValueError(f"Invalid parity '{parity}'. Must be one of {valid_parities}.")

    def _validate_bytesize(self, bytesize):
        if bytesize not in {7, 8}:
            raise ValueError(f"Invalid bytesize {bytesize}. Must be one of 7 or 8.")

    def _validate_protocol(self, protocol):
        valid_protocols = {'RS-232', 'RS-485'}
        if protocol not in valid_protocols:
            raise ValueError(f"Invalid protocol '{protocol}'. Must be one of {valid_protocols}.")
    #endregion

    #region CONFIGURATION
    def _configure_modbus_interface_type(self):
         # Send command to AC-422 to configure it for proper modbus communication
        parity_map = {'N': 0, 'O': 1, 'E': 2}
        protocol_map = {'RS-232': 0, 'RS-485': 1}

        cmd = '\x00\xff'.encode(self._encoding)
        cmd += struct.pack('>I', self._baudrate)
        cmd += chr(self._stopbits).encode(self._encoding)
        cmd += chr(parity_map.get(self._parity, 2)).encode(self._encoding)
        cmd += chr(protocol_map.get(self._protocol, 1)).encode(self._encoding)

        self._serial_device.write(cmd)
        response = self._serial_device.read()
        success = int.from_bytes(response)

        if not success:
            raise IOError("Error setting configuration for AC-422")
    #endregion
    
    #region CMD
    def _build_modbus_command(self, device_address, function_code, start_address, num_registers, values = None):
        # Pack the Modbus request into a bytearray, calucate crc, append crc to end of command
        if values is not None:
            pass


        command = struct.pack('>B B H H', device_address, function_code, start_address, num_registers)
        crc = self._calculate_modbus_crc(command) 
        command += struct.pack('<H', crc)
        
        return command

    def _calculate_modbus_crc(self, msg):
        # Somehow calculates the crc for the modbus command
        # https://stackoverflow.com/questions/69369408/calculating-crc16-in-python-for-modbus
        crc = 0xFFFF
        for n in range(len(msg)):
            crc ^= msg[n]
            for i in range(8):
                if crc & 1:
                    crc >>= 1
                    crc ^= 0xA001
                else:
                    crc >>= 1
        return crc
    #endregion

    #region COMMANDS
    def write_to_registers(self,
            device_address: int, # Slave Address
            start_address: int,
            number_of_registers: int,
            values: list[Number],
            no_response_expected: bool = False
            ):
        """
        Write registers (code 0x16)

        :param device_address: Modbus slave ID
        :param start_address: Start address to write to
        :param number_of_registers: Number of registers to read
        :param values: List of values to write
        :param no_response_expected: (optional) The client will not expect a response to the request

        This function is used to write a block of contiguous registers.
        """
        if not type(device_address) == int:
            raise ValueError("device_address must be of type <int>")
        
        if not type(start_address) == int:
            raise ValueError("start_address must be of type <int>")
        
        if not type(number_of_registers) == int:
            raise ValueError("number_of_registers must be of type <int>")
        
        function_code = 16
        # TODO: Figure out writing
        cmd = self._build_modbus_command(device_address, function_code, start_address, values)
        self._send_cmd_and_await_response(cmd)
        
    def read_from_registers(self,
            device_address: int, # Slave Address
            start_address: int,
            number_of_registers: int,
            no_response_expected: bool = False
            ):
        """
        Read holding registers (code 0x03).

        :param device_address: Modbus slave ID 
        :param start_address: Start address to read from
        :param number_of_registers: Number of registers to read
        :param no_response_expected: (optional) The client will not expect a response to the request

        This function is used to read the contents of a contiguous block
        of holding registers in a remote device. The Request specifies the
        starting register address and the number of registers.
        """   
        if not type(device_address) == int:
            raise ValueError("device_address must be of type <int>")
        
        if not type(start_address) == int:
            raise ValueError("start_address must be of type <int>")
        
        if not type(number_of_registers) == int:
            raise ValueError("number_of_registers must be of type <int>")
        
        function_code = 3
        cmd = self._build_modbus_command(device_address, function_code, start_address, number_of_registers)
        self._send_cmd_and_await_response(cmd)

    def retrieve_sensor_value(self, 
            device_address: int, # Slave Address
            sensor_model: ModbusSensorModel, 
            register_name: str
            ):
        """
        :param device_address: Modbus slave ID
        :param sensor_model: First two letters of sensor model from the ModbusSensorModel enum 
        :param register_name: The name of the value. Must be in the Sensor_Register dictionary

        This function is used to retrieve the desired value without needing to
        know the holding register or how to convert the data to the proper format.
        """   
        if sensor_model not in Sensor_Registers:
            raise ValueError(f"Sensor model '{sensor_model.value}' not found.")

        sensor_data = self.register_dict[sensor_model]

        if register_name not in sensor_data:
            raise ValueError(f"Value '{register_name}' not found for sensor model '{sensor_model.value}'.")

        register = sensor_data[register_name]
        
        # TODO: Confirm endianness
        response = self.read_from_registers(device_address, register, 2, False)
        value = self.convert_bytes_to_float(response)

        return value
    #endregion 
    
    #region HELPERS
    def convert_bytes_to_float(self, bytestring, big_endian: bool = True):
        unpack_format = '>f' if big_endian else '<f'
        float_value = struct.unpack(unpack_format, bytestring)[0]
        return float_value
    #endregion


# import argparse
# from mypkg.bluetooth.commands import set_name as bluetooth_set_name
# from mypkg.usb.commands import set_name as usb_set_name
# from mypkg.modbus.commands import get_data as modbus_get_data
# from mypkg.sdi12.commands import query_address as sdi12_query_address

# def main():
#     # Top-level parser for 'mypkg'
#     parser = argparse.ArgumentParser(prog="mypkg")
#     subparsers = parser.add_subparsers(dest="protocol", help="Choose protocol")

#     # Bluetooth commands
#     bluetooth_parser = subparsers.add_parser("bluetooth", help="Bluetooth related commands")
#     bluetooth_subparsers = bluetooth_parser.add_subparsers(dest="command", help="Bluetooth commands")

#     bluetooth_set_name_parser = bluetooth_subparsers.add_parser("set-name", help="Set Bluetooth device name")
#     bluetooth_set_name_parser.add_argument("address", help="Bluetooth device address")
#     bluetooth_set_name_parser.add_argument("name", help="Name to set for the Bluetooth device")
    
#     # USB commands
#     usb_parser = subparsers.add_parser("usb", help="USB related commands")
#     usb_subparsers = usb_parser.add_subparsers(dest="command", help="USB commands")

#     usb_set_name_parser = usb_subparsers.add_parser("set-name", help="Set USB device name")
#     usb_set_name_parser.add_argument("address", help="USB device address")
#     usb_set_name_parser.add_argument("name", help="Name to set for the USB device")
    
#     # Modbus commands
#     modbus_parser = subparsers.add_parser("modbus", help="Modbus related commands")
#     modbus_subparsers = modbus_parser.add_subparsers(dest="command", help="Modbus commands")

#     modbus_get_data_parser = modbus_subparsers.add_parser("get-data", help="Get data from a Modbus device")
#     modbus_get_data_parser.add_argument("device_id", help="Modbus device ID")
    
#     # SDI-12 commands
#     sdi12_parser = subparsers.add_parser("sdi12", help="SDI-12 related commands")
#     sdi12_subparsers = sdi12_parser.add_subparsers(dest="command", help="SDI-12 commands")

#     sdi12_query_address_parser = sdi12_subparsers.add_parser("query-address", help="Query SDI-12 address")
#     sdi12_query_address_parser.add_argument("address", help="SDI-12 device address")

#     # Parse arguments
#     args = parser.parse_args()

#     # Handle Bluetooth set-name command
#     if args.protocol == "bluetooth" and args.command == "set-name":
#         bluetooth_set_name(args.address, args.name)

#     # Handle USB set-name command
#     elif args.protocol == "usb" and args.command == "set-name":
#         usb_set_name(args.address, args.name)

#     # Handle Modbus get-data command
#     elif args.protocol == "modbus" and args.command == "get-data":
#         modbus_get_data(args.device_id)

#     # Handle SDI-12 query-address command
#     elif args.protocol == "sdi12" and args.command == "query-address":
#         sdi12_query_address(args.address)

# if __name__ == "__main__":
#     main()



import bleak
import asyncio
import string
from dataclasses import dataclass, asdict
from enum import Enum
from typing import Dict

from apogee.bluetooth.SensorClasses import SENSOR_REGISTRY, ApogeeSensor
from apogee.tools.MessageHandler import LoggingMessages

#region DATA STRUCTURES
@dataclass
class AdvertisingInfo:
    sensor_id: str = ""
    alias: str = ""
    serial_number: str = ""
    type: str = ""
#endregion

class BleScanner:
    #region INIT
    def __init__(self):
        try:
            self._scanner = bleak.BleakScanner(detection_callback=self._did_receive_advertising_data)
            self._scanning = asyncio.Event()
            self._loop = asyncio.get_event_loop()
            
            self._discovered_sensors: Dict[str, Dict[str, str]] = {}

            self.min_duration: int = 3

        except Exception as e:
            raise IOError(e)
    #endregion

    #region GETTERS/SETTERS
    def set_min_scan_time(self, 
                          min_time: int):
        """
        Set minimum scan time before determining if all Apogee sensor packets have been found

        :param min_time: The duration of time in seconds to set the minimum scan time.
        """

        if not type(min_time) == int:
            raise ValueError("Minimum scan time must be of type <int>")
        
        self.min_duration = min_time
        LoggingMessages.info(f"Ble scanner min scan time set to: {min_time}")
    #endregion

    #region SCAN
    async def scan(self, 
                   duration: int = 10, 
                   end_early_if_stable: bool = False
                   ) -> Dict[str, Dict[str, str]]:
        """
        Scan for nearby Apogee sensors

        :param duration: (optional) The duration of time in seconds to scan for Apogee sensors
        :param end_early_if_stable: (optional) If determined that all nearby sensors are found and no more information is needed, end scan early.
                                                 This may still result in a false negative if the initial packet discovery of another sensor did not occur before the set minimum duration
        
        :return: A dictionary mapping MAC addresses (str) to a dictionary containing advertising information.
                The advertising information includes:
                    - sensor_id: The integer representation of the sensor's id number
                    - alias: The assigned alias
                    - serial_number: The serial number
                    - type: The type of sensor
        """

        try:
            # Start scanner
            await self._scanner.start()
            self._scanning.set()
            LoggingMessages.info(f"Starting Ble scan for {duration} seconds. End early if stable set to: {end_early_if_stable}")

            # Set the min and max time for scan duration
            min_time = self._loop.time() + self.min_duration
            max_time = self._loop.time() + duration
            
            # Check every second while scanner is still running
            while self._scanning.is_set():
                await asyncio.sleep(1.0) # Sleep for one second between checks

                # Ensure the min scan duraction has at least elapsed
                if (self._loop.time() > min_time):

                    # Check if scan can end early with no missing packets
                    if end_early_if_stable and self._no_missing_packets():
                        LoggingMessages.debug("Ble scan exiting early due to no new packets found")
                        self._scanning.clear()

                    # Check if scan has continued for max scan duration
                    if self._loop.time() > max_time:
                        LoggingMessages.debug("Ble scan max time elapsed")
                        self._scanning.clear()

        except bleak.exc.BleakDBusError as e:
            raise IOError(f"Error occurred during Ble Scan: {e}")
        
        except Exception as e:
            raise IOError(e)

        finally:
            await self._stop_scan()
            return {
                address: asdict(info) 
                for address, info in self._discovered_sensors.items()
            }

    async def _stop_scan(self):
        await self._scanner.stop()
        LoggingMessages.info("Ble scan stopped")
    #endregion

    #region HANDLE ADVERTISING DATA
    def _did_receive_advertising_data(self, device, advertisement_data):
        # Filter for Apogee's manufacturer ID: 1604 (0x0644)
        if 1604 in advertisement_data.manufacturer_data:
            self._did_discover_apogee_sensor(device, advertisement_data.manufacturer_data[1604])
        
    def _did_discover_apogee_sensor(self, sensor, manufacturer_data):
        LoggingMessages.debug(f"Ble scan - Apogee sensor discovered. {manufacturer_data}")
        address = sensor.address

        if address not in self._discovered_sensors:
            LoggingMessages.debug(f"Ble scan - Adding new sensor with address: {address}")
            self._discovered_sensors[address] = AdvertisingInfo(address)

        self._interpret_manufacturer_data(address, manufacturer_data)

    def _interpret_manufacturer_data(self, address, manufacturer_data):
        sensor = self._discovered_sensors[address]
        
        # Check which manufacturer data packet was received
        packet_contains_alias_data = self._is_alias_data(manufacturer_data)
        if packet_contains_alias_data:
            self._populate_alias_info(sensor, manufacturer_data)
        else:
            self._populate_sensor_details(sensor, manufacturer_data)
    
    def _populate_alias_info(self, sensor: AdvertisingInfo, manufacturer_data):
        LoggingMessages.debug(f"Ble scan - Packet contains alias data: {manufacturer_data}")

        sensor.alias = manufacturer_data.decode('utf-8')
    
    def _populate_sensor_details(self, sensor: AdvertisingInfo, manufacturer_data):
        LoggingMessages.debug(f"Ble scan - Packet contains sensor details data: {manufacturer_data}")

        sensor_id = int(manufacturer_data[-1])
        serial_number = int.from_bytes(manufacturer_data[:2], byteorder='little')

        sensor.type = self._get_sensor_type_from_id(sensor_id)
        sensor.sensor_id = sensor_id
        sensor.serial_number = serial_number
    #endregion

    #region HELPERS
    def _is_alias_data(self, bytes):
        # TODO: Confirm this works on other platforms like Mac, RPi
        try:
            # Check if all characters are printable (either ASCII or Unicode printable)
            decoded_str = bytes.decode("utf-8")
            is_alias = all(c in string.printable or ord(c) > 127 for c in decoded_str)
            return is_alias
        except (UnicodeDecodeError, ValueError) as e:
            return False

    def _no_missing_packets(self):
        for sensor in self._discovered_sensors.values():
            # Determine a packet is missing if info from one of the packets missing
            if (sensor.alias == "") or (sensor.serial_number == "") or (sensor.type == ""):
                return False
        
        return True
    
    def _get_sensor_type_from_id(self, id):
        sensor_class = SENSOR_REGISTRY.get(id, ApogeeSensor)
        return sensor_class.type
    #endregion



from apogee.bluetooth.Calculators import *

#region BLE SENSOR CLASSES
class ApogeeSensor:
    sensorID = 0
    model = ""
    type = ""
    data_labels = []

    @classmethod
    def calculate_data(cls, data):
        return data

class SP_110:
    sensorID = 1
    model = "SP-110"
    type = "Pyranometer"
    data_labels = ["Shortwave"]

    @classmethod
    def calculate_data(cls, data):
        return data

class SP_510:
    sensorID = 2
    model = "SP-510"
    type = "Thermopile Pyranometer"
    data_labels = ["Shortwave"]
        
    @classmethod
    def calculate_data(cls, data):
        return data

class SP_610:
    sensorID = 3
    model = "SP-610"
    type = "Thermopile Pyranometer (Downward)"
    data_labels = ["Shortwave"]
        
    @classmethod
    def calculate_data(cls, data):
        return data

class SQ_110:
    sensorID = 4
    model = "SQ-110"
    type = "Quantum (Solar)"
    data_labels = ["PPFD"]
        
    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        return calculated_data
    
class SQ_120:
    sensorID = 5
    model = "SQ-120"
    type = "Quantum (Electric)"
    data_labels = ["PPFD"]
        
    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        return calculated_data
    
class SQ_500:
    sensorID = 6
    model = "SQ-500"
    type = "Quantum (Full Spectrum)"
    data_labels = ["PPFD"]
        
    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        return calculated_data

class SL_510:
    sensorID = 7
    model = "SL-510"
    type = "Pyrgeometer"
    data_labels = ["Longwave", "Temperature"]
        
    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        temp = temp_calculator(calculated_data[1])
        calculated_data[1] = temp
        return calculated_data
    
class SL_610:
    sensorID = 8
    model = "SL-610"
    type = "Pyrgeometer (Downward)"
    data_labels = ["Longwave", "Temperature"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        temp = temp_calculator(calculated_data[1])
        calculated_data[1] = temp

        return calculated_data


class SI_1XX:
    sensorID = 9
    model = "SI-1XX"
    type = "IR Sensor"
    data_labels = ["Target Temp", "Sensor Temp"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        target_temp = temp_calculator(calculated_data[0])
        sensor_temp = temp_calculator(calculated_data[1])

        calculated_data = [target_temp, sensor_temp]
        return calculated_data

class SU_200:
    sensorID = 10
    model = "SU-200"
    type = "UV Sensor"
    data_labels = ["UV_EFD", "UV_PFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        uv_pfd = uv_pfd_calculator(calculated_data[0])
        calculated_data.append(uv_pfd)

        return calculated_data

class SE_100:
    sensorID = 11
    model = "SE-100"
    type = "Photometric"
    data_labels = ["Illuminance"]

    @classmethod
    def calculate_data(cls, data):
        return data

class S2_111:
    sensorID = 12
    model = "S2-111"
    type = "NDVI"
    data_labels = ["red", "NIR"]

    @classmethod
    def calculate_data(cls, data):
        return data

class S2_112:
    sensorID = 13
    model = "S2-112"
    type = "NDVI (Downward)"
    data_labels = ["red", "NIR", "NDVI"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        ndvi = approx_ndvi_calculator(calculated_data[0], calculated_data[1])
        calculated_data.append(ndvi)

        return calculated_data

class S2_121:
    sensorID = 14
    model = "S2-121"
    type = "PRI"
    data_labels = ["green", "yellow"]

    @classmethod
    def calculate_data(cls, data):
        return data

class S2_122:
    sensorID = 15
    model = "S2-122"
    type = "PRI (Downward)"
    data_labels = ["green", "yellow"]

    @classmethod
    def calculate_data(cls, data):
        return data

class S2_131:
    sensorID = 16
    model = "S2-131"
    type = "Red/FarRed"
    data_labels = ["red", "far_red", "red_farRed_ratio", "percent_red_farRed"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        red_farRed_ratio = red_farRed_ratio_calculator(calculated_data[0], calculated_data[1])
        percent_red_farRed = percent_red_farRed_calculator(calculated_data[0], calculated_data[1])

        calculated_data.append(red_farRed_ratio)
        calculated_data.append(percent_red_farRed)

        return calculated_data

class S2_141:
    sensorID = 17
    model = "S2-141"
    type = "PAR/FAR"
    data_labels = ["PAR", "far_red", "percent_par_farRed", "Total PFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        percent_par_farRed = percent_par_farRed_calculator(calculated_data[0], calculated_data[1])
        total_pfd = total_pfd_calculator(calculated_data[0], calculated_data[1])
        
        calculated_data[0] = par
        calculated_data.append(percent_par_farRed)
        calculated_data.append(total_pfd)
        
        return calculated_data

class SQ_610:
    sensorID = 18
    model = "SQ-610"
    type = "ePAR"
    data_labels = ["ePPFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        
        return calculated_data

class ST_XX0:
    sensorID = 19
    model = "ST-XX0"
    type = "Thermistor"
    data_labels = ["Temperature"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        temp = temp_calculator(calculated_data[0])
        calculated_data = [temp]

        return calculated_data

class SP_700:
    sensorID = 20
    model = "SP-700"
    type = "Albedometer"
    data_labels = ["Upward", "Downward", "Albedo"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 2):
            return calculated_data
        
        albedo = albedo_calculator(calculated_data[0], calculated_data[1])
        calculated_data.append(albedo)

        return calculated_data

class SQ_620:
    sensorID = 21
    model = "SQ-620"
    type = "Quantum (Extended Range)"
    data_labels = ["PFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par

        return calculated_data

class SQ_640:
    sensorID = 22
    model = "SQ-640"
    type = "Quantum (Low-Light)"
    data_labels = ["PFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        
        return calculated_data

class NDVI_Pair:
    sensorID = 23
    model = "NDVI Pair"
    type = "Pair NDVI"
    data_labels = ["red", "NIR"]

    @classmethod
    def calculate_data(cls, data):
        if (len(data) != 4):
            return data
        
        red_reflectance = reflectance_calculator(data[2], data[0])
        nir_reflectance = reflectance_calculator(data[3], data[1])
        ndvi = pair_ndvi_calculator(red_reflectance, nir_reflectance)

        calculated_data = [ndvi, red_reflectance, nir_reflectance]
        return calculated_data

class PRI_Pair:
    sensorID = 24
    model = "PRI Pair"
    type = "Pair PRI"
    data_labels = ["green", "yellow"]

    @classmethod
    def calculate_data(cls, data):
        return [data[0], data[1]]

class AY_002:
    sensorID = 25
    model = "Four Channel"
    type = "Four Channel"
    data_labels = ["ch1", "ch2", "ch3", "ch4"]

    @classmethod
    def calculate_data(cls, data):
        return data

class AY_001:
    sensorID = 26
    model = "Dual Channel"
    type = "Dual Channel"
    data_labels = ["ch1", "ch2"]

    @classmethod
    def calculate_data(cls, data):
        return data

class SQ_100X:
    sensorID = 27
    model = "SQ-100X"
    type = "Quantum"
    data_labels = ["PPFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        return calculated_data

class SQ_313:
    sensorID = 28
    model = "SQ-313"
    type = "Line Quantum"
    data_labels = ["PPFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        calculated_data[0] = par
        return calculated_data

class SM_500:
    sensorID = 29
    model = "SM-500"
    type = "Guardian"
    data_labels = ["PPFD", "Temperature", "Humidity", "CO2", "Pressure", "Dew Point", "VPD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 5):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        temp = temp_calculator(calculated_data[1])
        dew_point = dew_point_calculator(calculated_data[2], calculated_data[1])
        vpd = vpd_calculator(calculated_data[2], calculated_data[1])

        calculated_data[0] = par
        calculated_data[1] = temp
        calculated_data.append(dew_point)
        calculated_data.append(vpd)
        
        return calculated_data

class SM_600:
    sensorID = 30
    model = "SM-600"
    type = "Guardian"
    data_labels = ["ePPFD", "Temperature", "Humidity", "CO2", "Pressure", "Dew Point", "VPD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 5):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        temp = temp_calculator(calculated_data[1])
        dew_point = dew_point_calculator(calculated_data[2], calculated_data[1])
        vpd = vpd_calculator(calculated_data[2], calculated_data[1])

        calculated_data[0] = par
        calculated_data[1] = temp
        calculated_data.append(dew_point)
        calculated_data.append(vpd)
        
        return calculated_data

class SM1HUH:
    sensorID = 31
    model = "SM1HUH"
    type = "Guardian"
    data_labels = ["PPFD", "Temperature", "Humidity", "CO2", "Pressure", "Dew Point", "VPD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 5):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        temp = temp_calculator(calculated_data[1])
        dew_point = dew_point_calculator(calculated_data[2], calculated_data[1])
        vpd = vpd_calculator(calculated_data[2], calculated_data[1])

        calculated_data[0] = par
        calculated_data[1] = temp
        calculated_data.append(dew_point)
        calculated_data.append(vpd)
        
        return calculated_data
    
class SM2HUH:
    sensorID = 32
    model = "SM2HUH"
    type = "Guardian"
    data_labels = ["PPFD", "Temperature", "Humidity", "CO2", "Pressure", "Dew Point", "VPD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 5):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        temp = temp_calculator(calculated_data[1])
        dew_point = dew_point_calculator(calculated_data[2], calculated_data[1])
        vpd = vpd_calculator(calculated_data[2], calculated_data[1])

        calculated_data[0] = par
        calculated_data[1] = temp
        calculated_data.append(dew_point)
        calculated_data.append(vpd)
        
        return calculated_data

class SM3HUH:
    sensorID = 33
    model = "SM3HUH"
    type = "Guardian"
    data_labels = ["PPFD", "Temperature", "Humidity", "CO2", "Pressure", "Dew Point", "VPD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 5):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        temp = temp_calculator(calculated_data[1])
        dew_point = dew_point_calculator(calculated_data[2], calculated_data[1])
        vpd = vpd_calculator(calculated_data[2], calculated_data[1])

        calculated_data[0] = par
        calculated_data[1] = temp
        calculated_data.append(dew_point)
        calculated_data.append(vpd)
        
        return calculated_data
    
class SM4HUH:
    sensorID = 34
    model = "SM4HUH"
    type = "Guardian"
    data_labels = ["PPFD", "Temperature", "Humidity", "CO2", "Pressure", "Dew Point", "VPD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 5):
            return calculated_data
        
        par = par_calculator(calculated_data[0])
        temp = temp_calculator(calculated_data[1])
        dew_point = dew_point_calculator(calculated_data[2], calculated_data[1])
        vpd = vpd_calculator(calculated_data[2], calculated_data[1])

        calculated_data[0] = par
        calculated_data[1] = temp
        calculated_data.append(dew_point)
        calculated_data.append(vpd)
        
        return calculated_data

class SO_100:
    sensorID = 35
    model = "SO-100"
    type = "Oxygen Sensor"
    data_labels = ["Oxygen"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = [data[0]]
        return calculated_data

class SO_200:
    sensorID = 36
    model = "SO-200"
    type = "Oxygen Sensor"
    data_labels = ["Oxygen"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = [data[0]]
        return calculated_data

class SU_300:
    sensorID = 37
    model = "SU-300"
    type = "UV Index"
    data_labels = ["UV_Index", "UBV_EFD", "UVB_PFD"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        uvb_efd = uvb_efd_calculator(calculated_data[0])
        uvb_pfd = uvb_pfd_calculator(calculated_data[0])

        calculated_data.append(uvb_efd)
        calculated_data.append(uvb_pfd)

        return calculated_data

class SF_110:
    sensorID = 38
    model = "SF-110"
    type = "Radiation Frost"
    data_labels = ["Temperature"]

    @classmethod
    def calculate_data(cls, data):
        calculated_data = data
        if (len(calculated_data) != 1):
            return calculated_data
        
        temp = temp_calculator(calculated_data[0])
        
        calculated_data = [temp]
        
        return calculated_data
#endregion

SENSOR_REGISTRY = {
    0: ApogeeSensor,
    1: SP_110,
    2: SP_510,
    3: SP_610,
    4: SQ_110,
    5: SQ_120,
    6: SQ_500,
    7: SL_510,
    8: SL_610,
    9: SI_1XX,
    10: SU_200,
    11: SE_100,
    12: S2_111,
    13: S2_112,
    14: S2_121,
    15: S2_122,
    16: S2_131,
    17: S2_141,
    18: SQ_610,
    19: ST_XX0,
    20: SP_700,
    21: SQ_620,
    22: SQ_640,
    23: NDVI_Pair,
    24: PRI_Pair,
    25: AY_002,
    26: AY_001,
    27: SQ_100X,
    28: SQ_313,
    29: SM_500,
    30: SM_600,
    31: SM1HUH,
    32: SM2HUH,
    33: SM3HUH,
    34: SM4HUH,
    35: SO_100,
    36: SO_200,
    37: SU_300,
    38: SF_110
}



from apogee.SerialClient import SerialClient

class SDI12Client(SerialClient):
    
    #region CONSTRUCTOR
    def __init__(self,
                 port: str,
                 baudrate: int = 1200,
                 timeout: int = 2,
                 auto_caps_cmd: bool = True
                 ):
        """
        :param port: comport name of connected serial device
        :param baudrate: (optional) 115200, 57600, 38400, 1920, 9600, 1200
        :param timeout: (optional) No response timeout in seconds
        :param auto_caps_cmd: (optional) Automatically capitalize command (excluding device address)
        """

        self._auto_capitalize_cmd = auto_caps_cmd

        super().__init__(port, baudrate, timeout)
    #endregion
    
    def _send_cmd(self, cmd):
        if self._auto_capitalize_cmd:
            cmd = cmd[0] + cmd[1:].upper()

        if cmd[-1] != '!':
            cmd += '!'

        print(cmd)
        command = (cmd).encode()
        print(command)
        super()._send_cmd(command)


from serial import Serial

class SerialClient():
    #region CONSTRUCTOR
    def __init__(self, port, baudrate, timeout):
        self._validate_port(port)
        self._validate_baudrate(baudrate)
        self._validate_timeout(timeout)

        self._port = port
        self._baudrate = baudrate
        self._timeout = timeout
        
        self._encoding = 'latin-1'
        self._connected = False
        self._serial_device = None
        
        self.open()
    
    def __del__(self):
        self.close()
    #endregion
    
    #region VALIDATION
    def _validate_baudrate(self, baudrate):
        valid_baudrates = {115200, 57600, 38400, 19200, 9600, 1200}
        if baudrate not in valid_baudrates:
            raise ValueError(f"Invalid baudrate {baudrate}. Must be one of {valid_baudrates}.")
    
    def _validate_timeout(self, timeout):
        if not isinstance(timeout, int):
            raise TypeError(f"Invalid timeout {timeout}. Must be an integer.")
        if timeout < 0:
            raise ValueError(f"Invalid timeout {timeout}. Must be a non-negative integer.")

    def _validate_port(self, port):
        if not isinstance(port, str):
            raise TypeError(f"Invalid port {port}. Must be a string.")    
    #endregion

    #region PROPERTIES
    @property
    def connected(self):
        return self._connected
    
    @connected.setter
    def connected(self, value):
        if self._connected != value:
            self._connected = value
    #endregion

    #region CONNECTION MANAGEMENT
    def open(self):
        try:
            self._serial_device = Serial(
                port=self._port,
                baudrate=self._baudrate,
                timeout=self._timeout
                )
            self.connected = True     
        except Exception as e:
            raise IOError(e) 
    
    def close(self):
        try:
            self._serial_device.close()
            self.connected = False
            self._serial_device = None
        except Exception as e:
            raise IOError(e)
    #endregion

    #region COMMUNICATION
    def _send_cmd(self, cmd):
        try:
            if self.connected:
                self._serial_device.flush()
                self._serial_device.write(cmd)
        except Exception as e:
            raise IOError(e)
            
    def _read_response(self):
        try:
            response = self._serial_device.readline()
            return response
        except Exception as e:
            raise IOError(e)
    
    def _send_cmd_and_await_response(self, cmd, no_response_expected = False):
        try:
            if self.connected:
                self._serial_device.reset_input_buffer()
                self._send_cmd(cmd)

                if not no_response_expected:
                    response = self._serial_device.readline()
                    return response
                else:
                    return ""
        except Exception as e:
            raise IOError(e)
    #endregion


import datetime
import struct
import uuid
import asyncio
import bleak
from bleak.backends.characteristic import BleakGATTCharacteristic
from typing import List, Dict, Callable, Tuple, Awaitable

from apogee.bluetooth.uuids import *
from apogee.bluetooth.SensorClasses import SENSOR_REGISTRY, ApogeeSensor
from apogee.tools.MessageHandler import LoggingMessages as messenger


class BleClient:
    #region INIT
    def __init__(self, 
                 address: str):
        """
        :param address: The MAC address of the sensor. If unknown, use the BleScanner to scan for and obtain the address
        """
        self._validate_address(address)
        
        self._address: str = address
        self._bleak_client: bleak.BleakClient = None 
        
        # notification_handlers used for live data notifications
        self.__notification_handlers: Dict[str, Callable] = {}

        messenger.debug(f"BleClient created with address: {address}")
    #endregion
    
    #region VALIDATION
    def _validate_address(self, address):
        if not isinstance(address, str):
            raise TypeError(f"Invalid address {address}. Must be a string.")  
    #endregion

    #region CONNECTION               
    async def connect(
            self,
            retries: int = 2,
            retry_delay: float = 0.5
            ) -> bool:
        """
        Attempt to connect to BLE device

        :param retries: (optional) An int representing the number of tries to retry in the event of a failed connection
        :param retry_delay: (optional) A float representing the number of seconds to wait between retries
        """
        if not self._bleak_client:
            exception_msg = "" # Save error message in the event of failuree

            for attempt in range(retries):
                messenger.info("Connecting to sensor...")
                try:
                    self._bleak_client = bleak.BleakClient(self._address)
                    await self._bleak_client.connect()
                    connected = self._bleak_client.is_connected
                    messenger.debug(f"Sensor connected: {connected}")
                    return connected
                except Exception as e:
                    messenger.info(f"Connection attempt {attempt + 1} failed")
                    exception_msg = e
                
                # Let the device sleep momentarily before retrying, unless it's the last attempt
                if attempt < retries - 1:
                    messenger.debug(f"Delaying for {retry_delay} sec before next attempt")
                    await asyncio.sleep(retry_delay)


            messenger.error("Failed to connect to sensor")
            raise RuntimeError(f"Failed to connect to sensor {self._address}. {exception_msg}")  

    async def disconnect(self):
        """
        Attempt to disconnect from BLE device
        """
        if self._bleak_client:
            try:
                messenger.info("Disconnecting from sensor")
                await self._bleak_client.disconnect()
                self._bleak_client = None
            
            except bleak.BleakError as e:
                raise RuntimeError(f"Error when disconnecting from sensor {self._address}. {e}")
    
    @property
    def is_connected(self) -> bool:
        """
        Return connected state of Ble device

        :return: A boolean indicating the current connected state of a Ble device
        :rtype: bool
        """
        
        if not self._bleak_client:
            return False
        
        connected = self._bleak_client.is_connected
        messenger.debug(f"Sensor connection state: {connected}")
        return connected
    #endregion

    #region GATT CHARACTERISTIC COMMUNICATION
    async def read_characteristic(
            self, 
            uuid: uuid.UUID
            ) -> bytes:
        """
        Reads a Bluetooth GATT characteristic by UUID.

        :param uuid: The UUID of the characteristic to read.
        :return: The raw bytes returned from the characteristic.
        :raises RuntimeError: If not connected to a device or the read fails.
        """
        
        if not self._bleak_client or not self._bleak_client.is_connected:
            raise RuntimeError("No connected bluetooth device")
        
        try:
            messenger.debug(f"Reading from characteristic {str(uuid)}")
            data = await self._bleak_client.read_gatt_char(uuid)
            return bytes(data)
        except bleak.exc.BleakCharacteristicNotFoundError as e:
            raise RuntimeError(f"Characteristic doesn't exist for connected sensor. {e}")
        except bleak.BleakError as e:
            raise RuntimeError(f"Read Failed: {e}")
        except Exception as e:
            raise RuntimeError(e)
    
    async def write_characteristic(
            self, 
            uuid: uuid.UUID, 
            data: bytearray
            ):
        """
        Writes a bytearray to a Bluetooth GATT characteristic by UUID.

        :param uuid: The UUID of the characteristic to write.
        :param data: The data to write of type bytearray.
        :raises RuntimeError: If not connected to a device or the write fails.
        """
        if not self._bleak_client or not self._bleak_client.is_connected:
            raise RuntimeError("No connected bluetooth device")
        
        try:
            messenger.debug(f"Writing to characteristic {str(uuid)}")
            await self._bleak_client.write_gatt_char(uuid, data, True)
        except bleak.exc.BleakCharacteristicNotFoundError as e:
            raise RuntimeError(f"Characteristic doesn't exist for connected sensor. {e}")
        except bleak.BleakError as e:
            raise RuntimeError(f"Write Failed: {e}")
        except Exception as e:
            raise RuntimeError(e)
        
    async def start_notifications(
            self, 
            uuid: uuid.UUID, 
            notification_handler: Callable[
                [BleakGATTCharacteristic, bytearray], 
                None | Awaitable[None]
                ]
            ):
        """
        Starts notifications of a Bluetooth GATT characteristic by UUID.

        :param uuid: The UUID of the characteristic to write.
        :param notification_handler: The callback function to receive the notifications.
        :raises RuntimeError: If not connected to a device or the notification start fails.
        """
        if not self._bleak_client or not self._bleak_client.is_connected:
            raise RuntimeError("No connected bluetooth device")
        
        try:
            messenger.debug(f"Starting notification for characteristic {str(uuid)}")
            await self._bleak_client.start_notify(uuid, notification_handler)
        except bleak.exc.BleakCharacteristicNotFoundError as e:
            raise RuntimeError(f"Characteristic doesn't exist for connected sensor. {e}")
        except bleak.BleakError as e:
            raise RuntimeError(f"Notification start failed: {e}")
        except Exception as e:
            raise RuntimeError(e)
        
    async def stop_notifications(
            self, 
            uuid: uuid.UUID
            ):
        """
        Starts notifications of a Bluetooth GATT characteristic by UUID.

        :param uuid: The UUID of the characteristic to write.
        :raises RuntimeError: If not connected to a device or the notification stop fails.
        """
        if not self._bleak_client or not self._bleak_client.is_connected:
            raise RuntimeError("No connected bluetooth device")
        
        try:
            messenger.debug(f"Stopping notification for characteristic {str(uuid)}")
            await self._bleak_client.stop_notify(uuid)
        except bleak.exc.BleakCharacteristicNotFoundError as e:
            raise RuntimeError(f"Characteristic doesn't exist for connected sensor. {e}")
        except bleak.BleakError as e:
            raise RuntimeError(f"Notification stop failed: {e}")
        except Exception as e:
            raise RuntimeError(e)
    #endregion

    #region BATTERY LEVEL
    async def get_battery_level(self) -> int:
        """
        Get the battery level of the connected sensor.
        Not applicable to Guardian sensors.
        
        :return: An int indicating the current battery level percentage.
        """
        data = await self.read_characteristic(batteryLevelCharacteristicUUID)
        messenger.debug(f"Battery level data read: {data}")
        battery = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Battery level read: {battery}")

        return battery
    #endregion

    #region DEVICE INFO
    async def get_device_info(self) -> Dict:
        """
        Get the basic device info of the connected sensor.

        :return: A dict containing the device information, including:
                - address: MAC address
                - manufacturer: Manufacturer Name
                - model: Model Number
                - serial number: Serial Number
                - firmware version: Firmware Version
                - hardware version: Hardware Version
                - software version: Software Version
                - alias: Device Alias
                - sensor id: Sensor ID number
        """
        messenger.info("Collecting device info...")

        manufacturer = await self.get_manufacturer()
        model = await self.get_model()
        serial_number = await self.get_serial_number()
        firmware = await self.get_firmware_version()
        hardware = await self.get_hardware_version()
        software = await self.get_software_version()
        alias = await self.get_alias()
        id = await self.get_sensor_id()

        return {
            "address": self._address,
            "manufacturer": manufacturer, 
            "model": model, 
            "serial number": serial_number, 
            "firmware version": firmware, 
            "hardware version": hardware, 
            "software version": software, 
            "alias": alias,
            "sensor id": id
            }
    #endregion

    #region MANUFACTURER NAME
    async def get_manufacturer(self) -> str:
        """
        Get the manufacturer name of the connected sensor
        
        :return: A str with the name of the sensor manufacturer.
        """
        data = await self.read_characteristic(manufacturerNameUUID)
        messenger.debug(f"Manufacturer name data read: {data}")

        manufacturer = data.decode('utf-8')
        messenger.info(f"Manufacturer name read: {manufacturer}")

        return manufacturer
    #endregion

    #region MODEL
    async def get_model(self) -> str:
        """
        Get the model number of the connected sensor
        
        :return: A str with the model number of the sensor.
        """
        data = await self.read_characteristic(modelNumberUUID)
        messenger.debug(f"Model number data read: {data}")

        model = data.decode('utf-8')
        messenger.info(f"Model number read: {model}")

        return model
    #endregion

    #region SERIAL NUMBER
    async def get_serial_number(self) -> int:
        """
        Get the serial number of the connected sensor
        
        :return: An int with the serial number of the sensor.
        """
        data = await self.read_characteristic(serialNumberUUID)
        messenger.debug(f"Serial number data read: {data}")

        serial_no = int(data.decode('utf-8'))
        messenger.info(f"Serial number read: {serial_no}")

        return serial_no
    #endregion

    #region FIRMWARE VERSION
    async def get_firmware_version(self) -> int:
        """
        Get the firmware version of the connected sensor
        
        :return: An int with the firmware version of the sensor.
        """
        data = await self.read_characteristic(firmwareVersionUUID)
        messenger.debug(f"Firmware version data read: {data}")

        fw = int(data.decode('utf-8'))
        messenger.info(f"Firmware version read: {fw}")

        return fw
    #endregion

    #region HARDWARE VERSION
    async def get_hardware_version(self) -> int:
        """
        Get the hardware version of the connected sensor
        
        :return: An int with the hardware version of the sensor.
        """
        data = await self.read_characteristic(hardwareVersionUUID)
        messenger.debug(f"Hardware version data read: {data}")

        hw = int(data.decode('utf-8'))
        messenger.info(f"Hardware version read: {hw}")

        return hw
    #endregion

    #region SOFTWARE VERSION
    async def get_software_version(self) -> str:
        """
        Get the software version of the connected sensor.
        A string with a single space (" ") denotes a release version.
        
        :return: A str with the software version of the sensor. 
        """
        data = await self.read_characteristic(softwareVersionUUID)
        messenger.debug(f"Software version data read: {data}")

        sw = data.decode('utf-8')
        messenger.info(f"Software version read: {sw}")

        return sw
    #endregion

    #region ALIAS
    async def get_alias(self) -> str:
        """
        Get the alias of the connected sensor.
        
        :return: A str with the alias of the sensor. 
        """
        data = await self.read_characteristic(aliasUUID)
        messenger.debug(f"Alias data read: {data}")

        alias = data.decode('utf-8')
        messenger.info(f"Alias read: {alias}")

        return alias
    
    async def set_alias(
            self, 
            alias: str):
        """
        Set the alias of the connected sensor.
        Max number of characters for alias is 20.
        
        :param alias: A str with the desired alias of the sensor. 
        """
        if not isinstance(alias, str):
            raise ValueError("Alias must be of type <str>")
        
        if len(alias) > 20:
            raise ValueError("Alias length has a maximum length of 20")
        
        messenger.info(f"Alias write: {alias}")

        data = bytearray(alias.encode('utf-8'))
        messenger.debug(f"Alias data write: {data}")

        await self.write_characteristic(aliasUUID, data)
    #endregion

    #region SENSOR ID
    async def get_sensor_id(self) -> int:
        """
        Get the sensor id of the connected sensor.
        
        :return: An int with the sensor id number. 
        """
        data = await self.read_characteristic(sensorIDUUID)
        messenger.debug(f"Sensor ID data read: {data}")

        id = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Sensor ID read: {id}")

        return id
    
    async def set_sensor_id(
            self, 
            id: int):
        """
        Change the sensor id of the connected sensor.
        This will also result in a change in the model of the sensor and the data recorded and returned from the sensor
        
        :param id: An int with the sensor id number. 
        """
        if not isinstance(id, int):
            raise ValueError("Sensor ID must be of type <int>")
        
        messenger.info(f"Sensor ID write: {id}")

        data = bytearray(id.to_bytes(1, byteorder='little', signed=False))
        messenger.debug(f"Sensor ID data write: {data}")

        await self.write_characteristic(sensorIDUUID, data)
    #endregion   
    
    #region LIVE DATA
    #TODO: Need to verify that these functions all work
    #TODO: Record an example of how to interact with this function that involve notifications
    async def start_live_data_notifications(self, notification_callback):
        """
        Start live data notifications

        :param uuid: uuid of the Gatt characteristic
        :param notification_handler: callable function that receives the decoded live data. The callback should accept the device address, sending uuid, and a list of the live data
        """
        if not callable(notification_callback):
            raise ValueError("The notification_callback must be a callable function.")
    
        uuid_str = str(liveDataUUID)
        self.__notification_handlers[uuid_str] = notification_callback

        await self.start_notifications(liveDataUUID, self._live_data_notifications_handler)

    async def stop_live_data_notifications(self):    
        """
        Stop live data notifications
        """    
        uuid_str = str(liveDataUUID)
        if uuid_str in self.__notification_handlers:
            del self.__notification_handlers[uuid_str]

        await self.stop_notifications(liveDataUUID)

    async def _live_data_notifications_handler(self, sender, data):
        uuid = sender.uuid
        live_data = self._interpret_live_data(data)

        if uuid in self.__notification_handlers:
            self.__notification_handlers[uuid](self._address, uuid, live_data)

    def _interpret_live_data(self, data) -> List[float]:
        values = []
        for i in range(0, len(data), 4):
            if len(data) < i + 4:
                break
            raw = int.from_bytes(data[i:i+4], byteorder='little', signed=True)
            value = raw / 10000.0  #  Divide by 10,000 to scale from ten-thousandths to ones
            values.append(value)
        
        return values
    #endregion

    #region LIVE DATA AVERAGING
    async def get_live_data_averaging_time(self) -> float:
        """
        Get the rolling live data average time in seconds of the connected sensor.
        
        :return: A float with the live data averaging time. 
        """
        data = await self.read_characteristic(liveDataAverageUUID)
        messenger.debug(f"Live data average data read: {data}")

        raw = int.from_bytes(data, byteorder='little', signed=False)
        average = float(raw) / 4.0 # Comes in units of 0.25 seconds from sensor
        messenger.info(f"Live data average read: {average}")

        return average
    
    async def set_live_data_averaging_time(
            self, 
            average: float):
        """
        Set the rolling live data average time in seconds of the connected sensor.
        
        :param average: A float with the number of seconds for the desired averaging time.
                        Averaging time must be between 0 and 30 seconds, inclusive 
        """
        try:
            avg = float(average)
        except (TypeError, ValueError):
            raise ValueError("Averaging value must be a number")

        if not (0 <= avg <= 30):
            raise ValueError("Averaging value must be 0 - 30, inclusive")
        
        messenger.info(f"Live data average write: {average}")

        value = int(avg * 4) # Convert to int in units of 0.25 seconds for the sensor
        data = bytearray(value.to_bytes(1, byteorder='little', signed=False))
        messenger.debug(f"Live data average data write: {data}")

        await self.write_characteristic(liveDataAverageUUID, data)
    #endregion

    #region LED STATE
    async def get_led_enabled(self) -> bool:
        """
        Get the led state of the connected sensor.
        Only applies to Guardian sensors.
        
        :return: A boolean indicating whether the LEDs are currently enabled.
        """
        data = await self.read_characteristic(LEDStateUUID)
        messenger.debug(f"LED state data read: {data}")

        state = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"LED state read: {state}")

        return bool(state)
    
    async def set_led_enabled(
            self, 
            enabled: bool):
        """
        Set the led state of the connected sensor.
        
        :param enabled: A boolean with the desired led state.
        """
        if not isinstance(enabled, bool):
            raise ValueError("Enabled must be of type <bool>")
        
        messenger.info(f"LED enabled write: {enabled}")

        state = int(enabled)
        data = bytearray(state.to_bytes(1, byteorder='little', signed=False))
        messenger.debug(f"LED enabled data write: {data}")

        await self.write_characteristic(LEDStateUUID, data)
    #endregion

    #region FAN CONTROL
    async def get_fan_state(self) -> Tuple[int, float, int, int]:
        """
        Get the fan state of the connected sensor.
        Only applies to Guardian sensors.

        :return: A tuple containing:
            - duty_cycle (int): The fan's duty cycle percentage.
            - fan_darkness_threshold (float): The darkness threshold value at which fan's duty cycle will be throttled. 0 indicates that the fan will never be throttled.
            - fan_pause_time (int): Time in minutes the fan is paused.
            - rpm (int): Fan rotations per minute.
        """
        data = await self.read_characteristic(fanControlUUID)
        messenger.debug(f"Fan control state data read: {data}")

        duty_cycle = int(data[0])
        fan_darkness_threshold = float(int.from_bytes(data[1:3], byteorder='little', signed=False) / 10)  # Divide by 10 to convert to float
        fan_pause_time = int(data[3])
        rpm = int.from_bytes(data[4:6], byteorder='little', signed=False)
        messenger.info(f"Fan control state read: {(duty_cycle, fan_darkness_threshold, fan_pause_time, rpm)}")

        return (duty_cycle, fan_darkness_threshold, fan_pause_time, rpm)
    
    async def set_fan_duty_cycle(
            self, 
            duty_cycle: int):
        """
        Set the duty cycle for the fan of the connected sensor.
        Only applies to Guardian sensors.
        
        :param duty_cycle: A float representing the desired duty cycle percentage of the fan.
                        duty_cycle must be between 40 and 100, inclusive 
        """
        if not isinstance(duty_cycle, int):
            raise ValueError("Duty cycle must be of type <int>")
        
        if not (40 <= duty_cycle <= 100):
            raise ValueError("Duty cycle must be 40 - 100, inclusive")
        
        messenger.info(f"Fan duty cycle write: {duty_cycle}")
        
        header = 1 # "1" is the header for updating just fan duty cycle
        data = bytearray(header.to_bytes(1, byteorder='little', signed=False)) 
        data.extend(bytearray(duty_cycle.to_bytes(1, byteorder='little', signed=False)))
        messenger.debug(f"Fan duty cycle data write: {data}")

        await self.write_characteristic(fanControlUUID, data)

        
    async def set_fan_pause_time(self, pause_time: int):
        """
        Set the number of minutes to pause the fan of the connected sensor.
        Only applies to Guardian sensors.
        
        :param pause_time: An int representing the number of minutes to pause the fan.
                        pause_time must be between 0 and 250 minutes, inclusive 
        """
        if not isinstance(pause_time, int):
            raise ValueError("Pause time must be of type <int>")
        
        if not (0 <= pause_time <= 250):
            raise ValueError("Pause time must be 0 - 250, inclusive")
        
        messenger.info(f"Fan pause time write: {pause_time}")
        
        header = 4 # "4" is the header for updating just fan pause time
        data = bytearray(header.to_bytes(1, byteorder='little', signed=False)) 
        data.extend(bytearray(pause_time.to_bytes(1, byteorder='little', signed=False)))
        messenger.debug(f"Fan pause time data write: {data}")

        await self.write_characteristic(fanControlUUID, data)    

    async def set_fan_darkness_threshold(self, darkness_threshold: float):
        """
        Set the darkness threshold for the fan of the connected sensor.
        When PAR values fall below this threshold, the fan's duty cycle will be throttled to 40%.
        Only applies to Guardian sensors.
        
        :param darkness_threshold: A float representing the PAR value threshold at which to throttle the fan.
                            darkness_threshold must be between 0 and 250 minutes, inclusive 
        """
        try:
            threshold = float(darkness_threshold)
        except (TypeError, ValueError):
            raise ValueError("Darkness threshold must be a number")
        
        if not (0 <= threshold <= 6553.5):
            raise ValueError("Darkness threshold must be 0 - 6553.5, inclusive")
        
        messenger.info(f"Fan darkness threshold write: {darkness_threshold}")
        
        darkness_threshold_converted = int(threshold * 10)  # Multiply by 10 to convert to int for bluetooth characteristic write
        
        header = 2 # "2" is the header for updating power saving mode
        data = bytearray(header.to_bytes(1, byteorder='little', signed=False)) 
        data.extend(bytearray(darkness_threshold_converted.to_bytes(2, byteorder='little', signed=False)))
        messenger.debug(f"Fan darkness threshold data write: {data}")

        await self.write_characteristic(fanControlUUID, data)
    #endregion

    #region DATA LOGGING
    async def start_data_logging(
            self,
            sampling_interval = 15, # seconds
            logging_interval = 300, # seconds 
            start_time = None, 
            end_time = None, 
            gateway_mode_rate = 0):
        """
        Turn on data logging, set logging parameters, and optionally turn on gateway mode in a single function call.
        
        :param sampling_interval (int): The number of seconds a data sample is taken.
        :param logging_interval (int): The number of seconds a data sample is logged.
        :param start_time (int): (optional) The time using epoch time in seconds at which the sensor did/will start logging. Is None if not logging and not set or not available on firmware version.
        :param end_tme (int): (optional) (firmware version restriction) The time using epoch time in seconds at which the sensor will end logging. Is None if not logging and not set or not available on firmware version.
        :param gateway_mode_rate (int)s: An int indicating how often the sensor will advertise when data is logged. 
                i.e., it advertises once every n data logs, where n correlates with the rate variable
                0 indicates that the sensor will not automatically advertise and requires a manual button press.
        """
            
        await self.set_logging_timing(sampling_interval, logging_interval, start_time, end_time)
        await self.set_logging_active(True)
        await self.set_gateway_mode_rate(gateway_mode_rate)

    async def get_logging_active(self) -> bool:
        """
        Get the logging state of the connected sensor.
        
        :return: A boolean indicating whether the sensor is currently logging.
        """
        data = await self.read_characteristic(dataLoggingControlUUID)
        messenger.debug(f"Data logging control data read: {data}")

        state = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Data logging control read: {state}")

        return bool(state)
    
    async def set_logging_active(
            self, 
            active: bool):
        """
        Set the logging state of the connected sensor.
        
        :param active: A boolean with the desired logging state.
        """
        if not isinstance(active, bool):
            raise ValueError("Enabled must be of type <bool>")
        
        messenger.info(f"Data logging control write: {active}")
        
        state = int(active)
        data = bytearray(state.to_bytes(1, byteorder='little', signed=False))
        messenger.debug(f"Data logging control data write: {data}")

        await self.write_characteristic(dataLoggingControlUUID, data)

    async def get_logging_timing(self) -> Tuple[int, int, int, int]:
        """
        Get the logging timing of the connected sensor.

        Firmware version restriction: End time is only available on uCache firmware version ≥ 9 or Guardian firmware version ≥ 3

        :return: A tuple containing:
            - sampling_interval (int): The number of seconds a data sample is taken.
            - logging_interval (int): The number of seconds a data sample is logged.
            - start_time (int): (optional) The time using epoch time in seconds at which the sensor did/will start logging. Is None if not logging and not set or not available on firmware version.
            - end_tme (int): (optional) (firmware version restriction) The time using epoch time in seconds at which the sensor will end logging. Is None if not logging and not set or not available on firmware version.
        """
        data = await self.read_characteristic(dataLoggingIntervalsUUID)
        messenger.debug(f"Data logging timing data read: {data}")

        sampling_interval = int.from_bytes(data[0:4], byteorder='little', signed=False)
        logging_interval = int.from_bytes(data[4:8], byteorder='little', signed=False)

        start_time = None
        end_time = None
        if len(data) >= 12:
            start_time = int.from_bytes(data[8:12], byteorder='little', signed=False)
        if len(data) >= 16:
            end_time = int.from_bytes(data[12:416], byteorder='little', signed=False)
        messenger.info(f"Data logging timing read: {(sampling_interval, logging_interval, start_time, end_time)}")

        return (sampling_interval, logging_interval, start_time, end_time)
    
    async def set_logging_timing(
            self, 
            sampling_interval: int, 
            logging_interval: int, 
            start_time: int = None, 
            end_time: int = None):
        """
        Set the logging timing of the connected sensor.

        Firmware version restriction: End time is only available on uCache firmware version ≥ 9 or Guardian firmware version ≥ 3

        :param sampling_interval: An int representing the number of seconds a data sample should be taken. MUST be less than AND an interval of the logging interval variable.
        :param logging_interval: An int representing the number of seconds a data sample should be logged. MUST be greater than AND divisible by the sampling interval variable.
        :param start_time: (optional) An int representing the desired start time in epoch time for data logging. Only available on firmware version...
        :param end_tme: (optional) (firmware version restriction) An int representing the desired end time in epoch time for data logging.  Only available on firmware version...
        """
        if not isinstance(sampling_interval, int):
            raise ValueError("Sampling interval must be of type <int>")
        if not isinstance(logging_interval, int):
            raise ValueError("Logging interval must be of type <int>")
        if not (isinstance(start_time, int) or start_time is None):
            raise ValueError("Start time must be of type <int> or None")
        if not (isinstance(end_time, int) or end_time is None):
            raise ValueError("End time must be of type <int> or None")
        if sampling_interval > logging_interval or logging_interval % sampling_interval != 0:
            raise ValueError("Sampling interval must be less than logging interval and be a factor of the logging interval")
        if (start_time != None and end_time != None and start_time >= end_time):
            raise ValueError("End time must be after start time")
        if (end_time != None and end_time < int(datetime.datetime.now().timestamp())):
            raise ValueError("End time must be after the current time")
        
        messenger.info(f"Data logging control write: {(sampling_interval, logging_interval, start_time, end_time)}")

        data = bytearray(sampling_interval.to_bytes(4, byteorder='little', signed=False))
        data.extend(bytearray(logging_interval.to_bytes(4, byteorder='little', signed=False)))

        if start_time and end_time: 
            data.extend(bytearray(start_time.to_bytes(4, byteorder='little', signed=False)))
            data.extend(bytearray(end_time.to_bytes(4, byteorder='little', signed=False)))
        elif end_time:
            no_start_time = 0 # Entering an end time requires something be entered for start time as well
            data.extend(bytearray(no_start_time.to_bytes(4, byteorder='little', signed=False)))
            data.extend(bytearray(end_time.to_bytes(4, byteorder='little', signed=False)))
        elif start_time:
            data.extend(bytearray(start_time.to_bytes(4, byteorder='little', signed=False)))
        messenger.debug(f"Data logging control data write: {data}")

        await self.write_characteristic(dataLoggingIntervalsUUID, data)
    #endregion
    
    #region GATEWAY MODE
    async def get_gateway_mode_rate(self) -> int:
        """
        Get the gateway mode rate of the connected sensor.
        Not applicable to Guardian sensors.

        Gateway mode provides the functionality for the sensor to advertise periodically, synchronized with data logging.
        The sensor will advertise for up to 10 seconds or until connected.
        
        :return: An int indicating how often the sensor will advertise when data is logged. 
                i.e., it advertises once every n data logs.
                0 indicates that the sensor will not automatically advertise and requires a manual button press.
        """
        data = await self.read_characteristic(dataLogCollectionRateUUID)
        messenger.debug(f"Gateway mode data read: {data}")

        rate = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Gateway mode read: {rate}")

        return rate
    
    async def set_gateway_mode_rate(
            self, 
            rate: int):
        """
        Set the gateway mode rate of the connected sensor.
        Not applicable to Guardian sensors.

        Gateway mode provides the functionality for the sensor to advertise periodically, synchronized with data logging.
        The sensor will advertise for up to 10 seconds or until connected.
        
        :param rate: An int indicating how often the sensor will advertise when data is logged. 
                i.e., it advertises once every n data logs, where n correlates with the rate variable
                0 indicates that the sensor will not automatically advertise and requires a manual button press.
        """
        if not isinstance(rate, int):
            raise ValueError("Rate must be of type <int>")
        
        if rate < 0:
            raise ValueError("Rate may not be negative")
        
        messenger.info(f"Gateway mode write: {rate}")
        
        data = bytearray(rate.to_bytes(1, byteorder='little', signed=False))
        messenger.debug(f"Gateway mode data write: {data}")

        await self.write_characteristic(dataLogCollectionRateUUID, data)
    #endregion

    #region COEFFICIENTS
    async def get_coefficients(self) -> List[float]:
        """
        Get all the calibration coefficients of the connected sensor.
        The number of calibration coefficients varies by sensor but will be between 1 and 6
        
        :return: A List containing 6 floats for the 6 possible calibration coefficients
        """
        coefficients_1 = await self.get_coefficients_1()
        coefficients_2 = await self.get_coefficients_2()
        coefficients = coefficients_1 + coefficients_2
        return list(coefficients)

    async def get_coefficients_1(self) -> List[float]:
        """
        Get the first 3 calibration coefficients of the connected sensor.
        
        :return: A List containing 3 floats for the first 3 calibration coefficients
        """
        data = await self.read_characteristic(coefficient1UUID)
        messenger.debug(f"Coefficients 1-3 data read: {data}")

        coefficients = struct.unpack('<3f', bytes(data)) 
        messenger.info(f"Coefficients 1-3 read: {coefficients}")

        return list(coefficients)

    async def get_coefficients_2(self) -> List[float]:
        """
        Get the last 3 calibration coefficients of the connected sensor.
        
        :return: A List containing 3 floats for the last 3 calibration coefficients
        """
        data = await self.read_characteristic(coefficient2UUID)
        messenger.debug(f"Coefficients 4-6 data read: {data}")

        coefficients = struct.unpack('<3f', bytes(data)) 
        messenger.info(f"Coefficients 4-6 read: {coefficients}")

        return list(coefficients)
    
    async def set_coefficients(
            self, 
            coefficients: List[float]):
        """
        Set all calibration coefficients of the connected sensor.
        If the desired sensor uses less than 6 calibration coefficients, set the remaining values to 0
        
        :param coefficients: A list of 6 floats containing the calibration coefficients.
        """
        if len(coefficients) != 6:
            raise ValueError("Coefficients must contains exactly 6 values. Use set_coefficients_1 or set_coefficients_2 to set only one group of coefficients")

        coefficients_1 = coefficients[0:3]
        coefficients_2 = coefficients[3:6]

        await self.set_coefficients_1(coefficients_1)
        await self.set_coefficients_2(coefficients_2)

    async def set_coefficients_1(
            self, 
            coefficients: List[float]):
        """
        Set the first 3 calibration coefficients of the connected sensor.
        If the desired sensor uses less than the first 3 calibration coefficients, set the remaining values to 0
        
        :param coefficients: A list of 3 floats containing the first 3 calibration coefficients.
        """
        if len(coefficients) != 3:
            raise ValueError("Coefficients must contains exactly 3 values. Use set_coefficients to set all coefficients at once. Otherwise, fill unused elements with 0")
       
        messenger.info(f"Coefficients 1-3 write: {coefficients}")

        data = struct.pack('<3f', *coefficients) 
        messenger.debug(f"Coefficients 1-3 data write: {data}")

        await self.write_characteristic(coefficient1UUID, data)

    async def set_coefficients_2(
            self, 
            coefficients: List[float]):
        """
        Set the last 3 calibration coefficients of the connected sensor.
        If the desired sensor uses less than the last 3 calibration coefficients, set the remaining values to 0
        
        :param coefficients: A list of 3 floats containing the last 3 calibration coefficients.
        """
        if len(coefficients) != 3:
            raise ValueError("Coefficients must contains exactly 3 values. Use set_coefficients to set all coefficients at once. Otherwise, fill unused elements with 0")

        messenger.info(f"Coefficients 4-6 write: {coefficients}")

        data = struct.pack('<3f', *coefficients)
        messenger.debug(f"Coefficients 4-6 data write: {data}")

        await self.write_characteristic(coefficient2UUID, data)
    #endregion

    #region MODBUS SETTINGS
    async def get_modbus_settings(self) -> Tuple[int, int, str, int]:
        """
        Get the modbus settings of the connected sensor.
        Only applies to Guardian sensors.

        :return: A tuple containing:
            - address (int): An int representing the modbus address of the sensor. 1 - 247
            - baudrate (int): An int representing the baudrate of the sensor. 115200, 57600, 38400, 19200, 9600, or 1200
            - parity (str): A string containing a single character representing the parity of the sensor. 'N'one or 'E'ven.
            - stop_bits (int): An int representing the stop bits of the sensor. 1 or 2
        """
        data = await self.read_characteristic(modbusSettingsUUID)
        messenger.debug(f"Modbus settings data read: {data}")

        address = data[0]
        baudrate_int = data[1]
        parity = data[2]
        stop_bits = data[3]

        parity_mapping = {0: 'N', 2: 'E'}
        parity_str = parity_mapping[parity]

        baudrate_mapping = {0: 115200, 1: 57600, 2: 38400, 3: 19200, 4: 9600, 5: 1200}
        baudrate = baudrate_mapping[baudrate_int]

        messenger.info(f"Modbus settings read: {(address, baudrate, parity_str, stop_bits)}")

        return (address, baudrate, parity_str, stop_bits)
    
    async def set_modbus_settings(self, address: int, baudrate: int, parity: str, stop_bits: int):
        """
        Set the modbus settings of the connected sensor.
        Only applies to Guardian sensors.

        :param address: An int representing the modbus address of the sensor. Valid values are 1 - 247, inclusive.
        :param baudrate: An int representing the baudrate of the sensor. 115200, 57600, 38400, 19200, 9600, or 1200
        :param parity: A string containing a single character representing the parity of the sensor. 'N'one or 'E'ven.
        :param stop_bits: An int representing the stop bits of the sensor. 1 or 2
        """
        if not isinstance(address, int):
            raise ValueError("address must be of type <int>")
        
        if not (1 <= address <= 247):
            raise ValueError("address must be between 1 and 247, inclusive")
        
        messenger.info(f"Modbus settings write: {(address, baudrate, parity, stop_bits)}")
        
        valid_baudrates = {115200, 57600, 38400, 19200, 9600, 1200}
        if baudrate not in valid_baudrates:
            raise ValueError(f"Invalid baudrate {baudrate}. Must be one of {valid_baudrates}.")

        valid_parities = {'N', 'E'}
        if parity not in valid_parities:
            raise ValueError(f"Invalid parity '{parity}'. Must be one of {valid_parities}.")

        if stop_bits not in {1, 2}:
            raise ValueError(f"Invalid stopbits {stop_bits}. Must be 1 or 2.")

        baudrate_mapping = {115200: 0, 57600: 1, 38400: 2, 19200: 3, 9600: 4, 1200: 5}
        baudrate_int = baudrate_mapping[baudrate]

        parity_mapping = {'N': 0, 'E': 2}
        parity_int = parity_mapping[parity]

        data = bytearray(address.to_bytes(1, byteorder='little', signed=False))
        data.extend(bytearray(baudrate_int.to_bytes(1, byteorder='little', signed=False)))
        data.extend(bytearray(parity_int.to_bytes(1, byteorder='little', signed=False)))
        data.extend(bytearray(stop_bits.to_bytes(1, byteorder='little', signed=False)))
        messenger.debug(f"Modbus settings data write: {data}")

        await self.write_characteristic(modbusSettingsUUID, data)
    #endregion

    #region SENSOR TIME
    async def get_sensor_time(self) -> int:
        """
        Get the current time of the sensor in epoch time.

        :return: An int representing epoch time in seconds
        """
        data = await self.read_characteristic(currentTimeUUID)
        messenger.debug(f"Sensor time data read: {data}")

        time = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Sensor time read: {time}")

        return time
    
    async def set_sensor_time(
            self, 
            time: int):
        """
        Set the time of the sensor to desired time in epoch time.

        :param time: An int representing epoch time in seconds.
        """
        if not isinstance(time, int):
            raise ValueError("Time must be of type <int>")
        
        if time < 0:
            raise ValueError("Time may not be negative")
        
        messenger.info(f"Sensor time write: {time}")
        
        data = bytearray(time.to_bytes(4, byteorder='little', signed=False))
        messenger.debug(f"Sensor time data write: {data}")

        await self.write_characteristic(currentTimeUUID, data)

    async def set_sensor_time_to_current(self):
        """
        Set the time of the sensor to the current time.
        """
        current_time = int(datetime.datetime.now().timestamp())

        data = bytearray(current_time.to_bytes(4, byteorder='little', signed=False))
        await self.write_characteristic(currentTimeUUID, data)
    #endregion

    #region TIME TILL FULL
    async def get_logging_full_time(self) -> int:
        """
        Get the time at which logging will be full in epoch time.
        When full, the sensor will start overwriting the oldest logs in memory.

        :return: An int representing epoch time in seconds at which logging will be full
        """
        data = await self.read_characteristic(timeTillFullUUID)
        messenger.debug(f"Time till full data read: {data}")

        time = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Time till full read: {time}")

        return time
    #endregion

    #region LAST TRANSFERRED TIMESTAMP
    async def get_last_transferred_timestamp(self) -> int:
        """
        Get the time in epoch time of the last datalog collected using the data transfer characteristic.

        :return: An int representing epoch time in seconds of the last log collected
        """
        data = await self.read_characteristic(lastTransferredTimestampUUID)
        messenger.debug(f"Last timestamp transferred data read: {data}")

        time = int.from_bytes(data, byteorder='little', signed=False)
        messenger.info(f"Last timestamp transferred read: {time}")

        return time
    
    async def set_last_transferred_timestamp(self, 
                                             time: int):
        """
        Set the time in epoch time of the last datalog collected.
        This will cause the sensor to start sending data from this timestamp next time logs are collected.

        :param time: An int representing epoch time in seconds of the last log collected
        """
        if not isinstance(time, int):
            raise ValueError("Time must be of type <int>")
        
        if time < 0:
            raise ValueError("Time may not be negative")
        
        messenger.info(f"Last timestamp transferred write: {time}")
        
        data = bytearray(time.to_bytes(4, byteorder='little', signed=False))
        messenger.debug(f"Last timestamp transferred data write: {data}")
        
        await self.write_characteristic(lastTransferredTimestampUUID, data)
    #endregion

    #region DARK OFFSET
    async def get_dark_offset_calibration(self):
        # TODO: Finish implementing this
        data = await self.read_characteristic(darkOffsetUUID)

        dark_offset = int.from_bytes(data, byteorder='little', signed=False)

        return dark_offset

    async def reset_dark_offset_calibration(self):
        data = bytearray([0x00]) # Packet to stop using dark offset calibration
        await self.write_characteristic(darkOffsetUUID, data)

    async def start_dark_offset_calibration(self, notification_callback):
        if not callable(notification_callback):
            raise ValueError("The notification_callback must be a callable function.")
    
        uuid_str = str(darkOffsetUUID)
        self.__notification_handlers[uuid_str] = notification_callback

        await self.start_notifications(darkOffsetUUID, self._dark_offset_calibration_completed)
        await self._initiate_dark_offset_calibration()

    async def _stop_dark_offset_notifications(self):        
        uuid_str = str(darkOffsetUUID)
        if uuid_str in self.__notification_handlers:
            del self.__notification_handlers[uuid_str]

        await self.stop_notifications(darkOffsetUUID)

    async def _initiate_dark_offset_calibration(self):
        data = bytearray([0x03]) # Packet to start dark offset calibration process
        await self.write_characteristic(darkOffsetUUID, data)

    async def _dark_offset_calibration_completed(self, sender, data):
        uuid = sender.uuid
        # TODO: Finish implementing this
        if uuid in self.__notification_handlers:
            self.__notification_handlers[uuid](self._address, uuid)
        
        await self._stop_dark_offset_notifications()
    #endregion

    #region DATA TRANSFER
    async def collect_data(self, 
                           start_time: int = None, 
                           end_time: int = None, 
                           timeout: int = None
                           ) -> List[Dict]:
    
        """
        Collect logged data from the sensor

        :param start_time: (optional) The time, using epoch time in seconds, of the earliest logged data to collect. 
                            A value of None will start from the last log transferred. (See get_last_timestamp_transferred)
        :param end_time: (optional) The time, using epoch time in seconds, of the latest logged data to collect.
                            A value of None will end at last log on device.
        :param timeout: (optional) The maximum number of seconds to collect data before aborting.
                            A value of None will never timeout.

        :return List: Returns a list of dicts that include the timestamp and the datapoints
        """

        if not (isinstance(start_time, int) or start_time is None):
            raise ValueError("Start time must be of type <int> or None")
        if not (isinstance(end_time, int) or end_time is None):
            raise ValueError("End time must be of type <int> or None")
        if not (isinstance(timeout, int) or timeout is None):
            raise ValueError("Timeout must be of type <int> or None")
        if (start_time != None and end_time != None and start_time >= end_time):
            raise ValueError("End time must be after start time")
        
        if start_time:
            await self.set_last_transferred_timestamp(start_time)

        # Info for calculating data from the raw values
        sensor_id = await self.get_sensor_id() 
        sensor_class = SENSOR_REGISTRY.get(sensor_id, ApogeeSensor)
        data_labels = sensor_class.data_labels
        data_calculation = sensor_class.calculate_data

        collected_data = []
        data_transfer_finished = asyncio.Event()
        data_transfer_end_timestamp = end_time if end_time is not None else 4294967295 # Highest number for a 32-bit unsigned int
        
        def _interpret_data_from_transfer(address, uuid, raw_data):
            nonlocal collected_data, sensor_class, data_labels, data_calculation

            # Callback for just compiling all the data before sending it to the user
            if uuid != str(dataLogTransferUUID) or address != self._address:
                return
            
            data_packet = bytes(raw_data)

            # Check if data packet is incomplete or if it is just all "FF" indicating end of data
            if len(data_packet) < 8 or data_packet == b'\xff' * 4:
                messenger.debug(f"End of data or incomplete packet. {data_packet}")
                data_transfer_finished.set()
                return
            
            # Get packet header information
            timestamp = struct.unpack('<I', data_packet[:4])[0]
            interval_between_timestamps = struct.unpack('<H', data_packet[4:6])[0]
            measurements_per_interval = struct.unpack('<B', data_packet[6:7])[0]

            # Separate packet into groups based on timestamp (e.g., groups of 5 datapoints for the Guardian, groups of 1 datapoint for microcache)
            for start_index in range(8, len(data_packet) - 1, 4 * measurements_per_interval):
                if timestamp > data_transfer_end_timestamp:
                    messenger.debug(f"Data timestamp {timestamp} after desired timestamp range. Ending collection.")
                    data_transfer_finished.set()
                    return

                end_index = min(start_index + (4 * measurements_per_interval), len(data_packet))
                grouped_array = data_packet[start_index:end_index]

                # Get each datapoint within the current timestamp
                raw_data = []
                for i in range(0, len(grouped_array), 4):
                    raw = struct.unpack('<i', grouped_array[i:(i + 4)])[0]

                    # Divide by 10,000 to scale from ten-thousandths to ones
                    value = raw / 10000.0

                    raw_data.append(value)

                # Calculate data from raw data and append to collected_data
                values_dict = {}
                calculated_data = data_calculation(raw_data)
                values_dict['Timestamp'] = timestamp
                for i, value in enumerate(calculated_data):
                    key = data_labels[i] if i < len(data_labels) else f"Data_{i}"
                    values_dict[key] = value
                collected_data.append(values_dict)

                # Increment timestamp in case there are multiple logs in a single packet
                timestamp += interval_between_timestamps
        
        messenger.info(f"Collecting data. Start time: {start_time}, End time: {end_time}, Timeout: {timeout}")

        await self.start_data_transfer_notifications(_interpret_data_from_transfer)
        await asyncio.wait_for(data_transfer_finished.wait(), timeout=timeout)
        await self.stop_data_transfer_notifications()

        return collected_data
    
    async def start_data_transfer_notifications(
            self, 
            notification_callback: Callable[[str, str, bytes], None]
            ):
        """
        Starts data transfer notifications for connected sensor.
        *Recommended to use the collect_data function which handles notifications and callbacks automatically.

        :param notification_callback: The callback function to receive the notifications.
        """
        if not callable(notification_callback):
            raise ValueError("The notification_callback must be a callable function.")
    
        messenger.debug(f"Starting data transfer notifications")

        uuid_str = str(dataLogTransferUUID)
        self.__notification_handlers[uuid_str] = notification_callback

        await self.start_notifications(dataLogTransferUUID, self._data_transfer_notifications_handler)

    async def stop_data_transfer_notifications(self):    
        """
        Stops data transfer notifications for connected sensor.
        *Recommended to use the collect_data function which handles notifications and callbacks automatically.
        """    
        uuid_str = str(dataLogTransferUUID)
        if uuid_str in self.__notification_handlers:
            del self.__notification_handlers[uuid_str]

        messenger.debug(f"Stopping data transfer notifications")

        await self.stop_notifications(dataLogTransferUUID)

    async def _data_transfer_notifications_handler(self, sender, data):
        uuid = sender.uuid

        messenger.debug(f"Data transfer packet received: {data}")
        
        # Send the data packet to the callback handler
        if uuid in self.__notification_handlers:
            self.__notification_handlers[uuid](self._address, uuid, data)
    #endregion
    



from enum import Enum

class ModbusSensorModel(Enum):
    Albedo = "Albedo"
    SP = "SP"
    SQ = "SQ"
    S2 = "S2"
    SI = "SI"
    SN = "SN"
    SM = "SM"

Sensor_Registers = {
    ModbusSensorModel.Albedo : {
        "calibrated_output_incident": 0,
        "calibrated_output_reflected": 2,
        "incident_detector_mv": 4,
        "reflected_detector_mv": 6,
        "albedo": 8,
        "device_status": 12,
        "firmware_version": 14,
        "address": 20,
        "model": 22,
        "serial": 24,
        "baudrate": 26,
        "parity": 28,
        "stopbits": 30,
        "multiplier_incident": 32,
        "offset_incident": 34,
        "multiplier_reflected": 36,
        "offset_reflected": 38,
        "running_average": 40,
        "heater_status": 42
    },
    ModbusSensorModel.SP : {
        "calibrated_output": 0,
        "detector_mv": 2,
        "device_status": 6,
        "firmware_version": 8,
        "address": 16,
        "model": 18,
        "serial": 20,
        "baudrate": 22,
        "parity": 24,
        "stopbits": 26,
        "multiplier": 28,
        "offset": 30,
        "running_average": 32,
        "heater_status": 34
    },
    ModbusSensorModel.SQ : {
        "calibrated_output": 0,
        "detector_mv": 2,
        "immersed_output": 4,
        "device_status": 10,
        "firmware_version": 12,
        "address": 16,
        "model": 18,
        "serial": 20,
        "baudrate": 22,
        "parity": 24,
        "stopbits": 26,
        "multiplier": 28,
        "offset": 30,
        "immersion_factor": 32,
        "running_average": 36,
        "heater_status": 38
    },
    ModbusSensorModel.S2 : {
        "calibrated_output_PAR": 0,
        "calibrated_output_FAR": 2,
        "detector_mv_PAR": 4,
        "detector_mv_FAR": 6,
        "ratio": 8,
        "percentage_FAR": 10,
        "device_status": 14,
        "firmware_version": 16,
        "address": 20,
        "model": 22,
        "serial": 24,
        "baudrate": 26,
        "parity": 28,
        "stopbits": 30,
        "multiplier_PAR": 32,
        "offset_PAR": 34,
        "multiplier_FAR": 36,
        "offset_FAR": 38,
        "running_average": 40,
        "heater_status": 42
    },
    ModbusSensorModel.SI : {
        "target_temp": 0,
        "sensor_temp": 2,
        "detector_mv": 4,
        "thermistor_mv": 6,
        "device_status": 10,
        "firmware_version": 12,
        "address": 20,
        "model": 22,
        "serial": 24,
        "baudrate": 26,
        "parity": 28,
        "stopbits": 30,
        "mc0": 32,
        "mc1": 34,
        "mc2": 36,
        "bc0": 38,
        "bc1": 40,
        "bc2": 42,
        "running_average": 44
    },
    ModbusSensorModel.SN : {
        "calibrated_shortwave_up": 0,
        "calibrated_shortwave_down": 2,
        "calibrated_longwave_up": 4,
        "calibrated_longwave_down": 6,
        "shortwave_net": 8,
        "longwave_net": 10,
        "total_net": 12,
        "albedo": 14,
        "shortwave_up_mv": 16,
        "shortwave_down_mv": 18,
        "longwave_up_mv": 20,
        "longwave_down_mv": 22,
        "longwave_up_temp": 24,
        "longwave_down_temp": 26,
        "device_status": 30,
        "firmware_version": 12,
        "address": 40,
        "model": 42,
        "serial": 44,
        "baudrate": 46,
        "parity": 48,
        "stopbits": 50,
        "longwave_up_multiplier": 52,
        "longwave_up_offset": 54,
        "longwave_down_multiplier": 56,
        "longwave_down_offset": 58,
        "shortwave_up_multiplier": 60,
        "shortwave_down_multiplier": 62,
        "running_average": 64,
        "heater_status": 66
    },
    ModbusSensorModel.SM : {
        "PAR": 0,
        "temp": 2,
        "humidity": 4,
        "co2": 6,
        "pressure": 8,
        "vpd": 10,
        "dew_point": 12,
        "fan_rpm": 14,
        "address": 40,
        "model": 42,
        "serial": 44,
        "baudrate": 46,
        "parity": 48,
        "stopbits": 50,
        "device_status": 52,
        "firmware_version": 54,
        "factory_PAR_multiplier": 56,
        "PAR_multiplier": 58,
        "PAR_offset": 60,
        "temp_offset": 62,
        "humidity_offset": 64,
        "co2_offset": 66,
        "running_average": 68, 
        "bluetooth_control": 70, 
        "LED_control": 72, 
        "fan_duty_cycle": 74, 
        "darkness_threshold": 76, 
        "modbus_measurement_control": 78, 
    }
}


import logging

class MessageHandler:
    def __init__(self):
        self.logger = logging.getLogger("my_sdk")
        self.set_level("WARNING")  # default level

        ch = logging.StreamHandler()
        formatter = logging.Formatter("[%(asctime)s] [%(levelname)s] %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        self.logger.propagate = False

    def set_level(self, level):
        """
        Accepts either a logging level integer or string name, e.g. logging.INFO or "INFO".
        """
        if isinstance(level, int):
            self.logger.setLevel(level)
        elif isinstance(level, str):
            level = level.upper()
            if level in logging._nameToLevel:
                self.logger.setLevel(logging._nameToLevel[level])
            else:
                raise ValueError(f"Invalid log level string: {level}")
        else:
            raise TypeError("Level must be int or str")

    #region METHODS
    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(msg, *args, **kwargs)
    #endregion


# shared logger instance
LoggingMessages = MessageHandler()


import uuid

# Battery Service
batteryServiceUUID = uuid.UUID("0000180F-0000-1000-8000-00805f9b34fb")
batteryLevelCharacteristicUUID = uuid.UUID("00002A19-0000-1000-8000-00805f9b34fb")

# Device Information Service
deviceInformationServiceUUID = uuid.UUID("0000180a-0000-1000-8000-00805f9b34fb")
manufacturerNameUUID = uuid.UUID("00002A29-0000-1000-8000-00805F9B34FB")
modelNumberUUID = uuid.UUID("00002A24-0000-1000-8000-00805F9B34FB")
serialNumberUUID = uuid.UUID("00002A25-0000-1000-8000-00805F9B34FB")
firmwareVersionUUID = uuid.UUID("00002A26-0000-1000-8000-00805F9B34FB")
hardwareVersionUUID = uuid.UUID("00002A27-0000-1000-8000-00805F9B34FB")
softwareVersionUUID = uuid.UUID("00002A28-0000-1000-8000-00805F9B34FB")

# Apogee Service
apogeeServiceUUID = uuid.UUID("B3E00001-2594-42A1-A5FE-4E660FF2868F")
liveDataUUID = uuid.UUID("B3E00002-2594-42A1-A5FE-4E660FF2868F")
sensorIDUUID = uuid.UUID("B3E00003-2594-42A1-A5FE-4E660FF2868F")
aliasUUID = uuid.UUID("B3E00004-2594-42A1-A5FE-4E660FF2868F")
liveDataAverageUUID = uuid.UUID("B3E00005-2594-42A1-A5FE-4E660FF2868F")
LEDStateUUID = uuid.UUID("B3E00006-2594-42A1-A5FE-4E660FF2868F")
fanControlUUID = uuid.UUID("B3E00007-2594-42A1-A5FE-4E660FF2868F")
dataLoggingControlUUID = uuid.UUID("B3E00010-2594-42A1-A5FE-4E660FF2868F")
dataLoggingIntervalsUUID = uuid.UUID("B3E00012-2594-42A1-A5FE-4E660FF2868F")
dataLogTransferUUID = uuid.UUID("B3E00013-2594-42A1-A5FE-4E660FF2868F")
dataLogCollectionRateUUID = uuid.UUID("B3E00014-2594-42A1-A5FE-4E660FF2868F")
coefficient1UUID = uuid.UUID("B3E00100-2594-42A1-A5FE-4E660FF2868F")
coefficient2UUID = uuid.UUID("B3E00101-2594-42A1-A5FE-4E660FF2868F")
modbusSettingsUUID = uuid.UUID("B3E00200-2594-42A1-A5FE-4E660FF2868F")
currentTimeUUID = uuid.UUID("B3E0000A-2594-42A1-A5FE-4E660FF2868F")
timeTillFullUUID = uuid.UUID("B3E0000C-2594-42A1-A5FE-4E660FF2868F")
logEntriesUUID = uuid.UUID("B3E0000D-2594-42A1-A5FE-4E660FF2868F")
lastTransferredTimestampUUID = uuid.UUID("B3E0000E-2594-42A1-A5FE-4E660FF2868F")
darkOffsetUUID = uuid.UUID("B3E000FF-2594-42A1-A5FE-4E660FF2868F")


# For help with building, packaging, and uploading to PyPI, visit:
# https://packaging.python.org/en/latest/tutorials/packaging-projects/
# and
# https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/
#
# TLDR; For just building and uploading, `python -m build` and then `twine upload dist/*` (Get API token from Bitwarden vault)
#

from setuptools import setup

# # This dynamic version of getting version isn't working anymore
# def get_version(rel_path):
#     with open(rel_path) as f:
#         exec(f.read())
#     return locals()['__version__']

setup(
    # version=get_version("./ApogeeSDK/version.py")
)


import math

def par_calculator(par: float) -> float:
    # TODO: Add a setting for par filtering
    return par

def temp_calculator(tempC: float) -> float:
    #TODO: Add a setting for unit conversion
    return tempC

def dew_point_calculator(humidity: float, tempC: float) -> float:
    saturation = saturation_vapor_pressure_calculator(tempC)
    vapor_pressure = (humidity / 100.0) * saturation
    log_term = (vapor_pressure / 0.61121)
    log = math.log(log_term)
    dew_point =  (257.14 * log) / (18.678 - log)

    # TODO: Add a setting for unit conversion
    return round(dew_point, 3)

def vpd_calculator(humidity: float, tempC: float) -> float:
    saturation = saturation_vapor_pressure_calculator(tempC)
    vapor_pressure = (humidity / 100.0) * saturation
    vpd = saturation - vapor_pressure

    return round(vpd, 3)

def saturation_vapor_pressure_calculator(tempC: float) -> float:
    exponent = ((18.678 - (tempC / 234.5)) * tempC) / (257.14 + tempC)
    return (0.6121 * math.exp(exponent))

def uv_pfd_calculator(uv_efd: float) -> float:
    return uv_efd * 3.0

def uvb_efd_calculator(uvi: float) -> float:
    return -0.0037 * uvi ** 2 + 0.21 * uvi + 0.0

def uvb_pfd_calculator(uvi: float) -> float:
    uvb_efd = uvb_efd_calculator(uvi)
    return uvb_efd * 2.6

def approx_ndvi_calculator(red: float, nir: float) -> float:
    return ((1.35 * nir) - red) / ((1.35 * nir) + red)

def reflectance_calculator(radiance: float, irradiance: float) -> float:
    return radiance / irradiance

def pair_ndvi_calculator(red_reflectance: float, nir_reflectance: float) -> float:
    return (nir_reflectance - red_reflectance) / (nir_reflectance + red_reflectance)

def red_farRed_ratio_calculator(red: float, far_red: float) -> float:
    return red / far_red

def percent_red_farRed_calculator(red: float, far_red: float) -> float:
    return (far_red / (red + far_red) * 100)

def percent_par_farRed_calculator(par: float, far_red: float) -> float:
    return (far_red / (par + far_red) * 100)

def total_pfd_calculator(par: float, far_red: float) -> float:
    return par + far_red

def albedo_calculator(upward: float, downward: float) -> float:
    return downward / upward
